local bfs = {}

local move_offset = {
	vec2(0, -1),
	vec2(0, 1),
	vec2(-1, 0),
	vec2(1, 0)
}

local n = 100

function bfs.walkable(start_pos, is_walkable_fn, max_distance)
	local queue = { {start_pos, 0} }
	local walkable_tiles = {}
	local visited = {}

	while #queue > 0 do
		local pos = queue[1][1]
		local distance = queue[1][2]
		table.remove(queue, 1)

		if distance <= max_distance then
			table.insert(walkable_tiles, vec2(pos))

			for _, dir in ipairs(move_offset) do
				local new_pos = pos + dir
				local idx = new_pos.x + new_pos.y * n
				if not visited[idx] and is_walkable_fn(new_pos) then
					visited[idx] = true
					table.insert(queue, {new_pos, distance + 1})
				end
			end
		end
	end

	return walkable_tiles
end

function bfs.path(start_pos, end_pos, is_walkable_fn)
	local queue = { start_pos }
	local walkable_tiles = {}
	local visited = {}

	while #queue > 0 do
		local pos = queue[1]
		table.remove(queue, 1)

		if not eql_pos(pos, end_pos) then
			table.insert(walkable_tiles, vec2(pos))

			for _, dir in ipairs(move_offset) do
				local new_pos = pos + dir
				local idx = new_pos.x + new_pos.y * n
				if not visited[idx] and is_walkable_fn(new_pos) then
					visited[idx] = vec2(pos)
					table.insert(queue, new_pos)
				end
			end
		else
			break
		end
	end

	-- reconstruct path
	local path = {}
	local pos = end_pos

	if visited[pos.x + pos.y * n] ~= nil then
		while not eql_pos(pos, start_pos) do
			table.insert(path, pos)
			pos = visited[pos.x + pos.y * n]
		end
	end

	return path
end

function bfs.raycast(start_x, start_y, end_x, end_y, nopath, is_opaque_fn)
	local tiles = nil
	if not nopath then
		tiles = {}
	end

	-- simply do bresenheim's
	local dx = end_x - start_x
	local dy = end_y - start_y
	local e = 0
	local de = math.abs(dy / dx)
	local y = start_y

	local x_dir = 1
	if dx < 0 then
		x_dir = -1
	end

	local y_dir = 1
	if dy < 0 then
		y_dir = -1
	end

	local obj, res
	local last_x
	for x=start_x,end_x,x_dir do
		last_x = x
		if not nopath then
			table.insert(tiles, vec2(x, y))
		end
		if is_opaque_fn(vec2(x, y)) then
			break
		end
		e = e + de

		while e > 0.5 do
			e = e - 1
			y = y + y_dir

			if e > 0.5 then
				if not nopath then
					table.insert(tiles, vec2(x, y))
				end
				if is_opaque_fn(vec2(x, y)) then
					break
				end
			end

			if y == end_y and x == end_x then
				if e <= 0.5 and not nopath then
					table.insert(tiles, vec2(x, y))
				end
				break
			end
		end
	end

	local reached = not (last_x ~= end_x or y ~= end_y)
	
	if nopath then
		return not reached
	else
		return tiles, reached
	end
end

function bfs.visible(start_pos, is_opaque_fn, d)
	-- launch rays from center to dxd box around start_pos
	local visible = {}
	local function add_if_unique(pos)
		if not pos_in_table(pos, visible) then
			table.insert(visible, pos)
		end
	end

	for px=start_pos.x-d,start_pos.x+d do
		local py = start_pos.y-d
		local path = bfs.raycast(start_pos.x, start_pos.y, px, py, false, is_opaque_fn)
		for _,pos in ipairs(path) do
			add_if_unique(pos)
		end
	end
	for px=start_pos.x-d,start_pos.x+d do
		local py = start_pos.y+d
		local path = bfs.raycast(start_pos.x, start_pos.y, px, py, false, is_opaque_fn)
		for _,pos in ipairs(path) do
			add_if_unique(pos)
		end
	end
	for py=start_pos.y-(d-1),start_pos.y+(d-1) do
		local px = start_pos.x+d
		local path = bfs.raycast(start_pos.x, start_pos.y, px, py, false, is_opaque_fn)
		for _,pos in ipairs(path) do
			add_if_unique(pos)
		end
	end
	for py=start_pos.y-(d-1),start_pos.y+(d-1) do
		local px = start_pos.x-d
		local path = bfs.raycast(start_pos.x, start_pos.y, px, py, false, is_opaque_fn)
		for _,pos in ipairs(path) do
			add_if_unique(pos)
		end
	end

	return visible
end

return bfs
