local room = {}
local rules = require('rules')
local bfs = require('bfs')
local ai = require('ai')

local obj_layer = 1
local overlay_layer = 8 
local push_overlay_layer = 12

local width, height
local objects

local selected_unit

-- room is either in animation state or not,
local animation = false
local animation_len = 0.083
local animation_start = 0.0

local move_offset = {
	['up'] = vec2(0, -1),
	['down'] = vec2(0, 1),
	['left'] = vec2(-1, 0),
	['right'] = vec2(1, 0)
}

-- room has different states:
-- 'wait' - waiting for input
-- 'move' - unit is in move animation
-- 'enemy' - enemies are moving
local room_state = nil

local function world2screen(pos)
	return vec2(
		scr_size.x/2 + (pos.x - width/2) * 64 + 32,
		scr_size.y/2 + (pos.y - height/2) * 64 + 32
	)
end

local function screen2world(pos)
	local x = pos.x - scr_size.x/2 
	local y = pos.y - scr_size.y/2 
	x = math.floor(x / 64 + width/2)
	y = math.floor(y / 64 + height/2)
	if x >= 0 and x < width then
		if y >= 0 and y < height then
			return vec2(x, y)
		end
	end
	return nil
end

function room.for_each_unit(fn)
	for _,obj in ipairs(objects) do
		if obj.id == 'player' then
			fn(obj)
		end
	end
end

function room.for_each_unit_at_pos(pos, fn)
	room.for_each_unit(function(u)
		if eql_pos(pos, u.pos) then
			fn(u)
		end
	end)
end

function room.for_each_enemy(fn)
	for _,obj in ipairs(objects) do
		if obj.id == 'enemy' then
			fn(obj)
		end
	end
end

function room.init()
end

function room.close()
end

function room.preenter()
	objects, width, height = rules.parse_level(room.level)
	room.win = false
	room.exiting = false
	room.units_died = 0
end

-- returns nil or list of objs at pos
function room.find_at_pos(pos, at_next)
	local res = {}
	for i,obj in ipairs(objects) do
		local p = obj.pos
		if at_next and obj.next_pos then
			p = obj.next_pos
		end
		if feql(p.x, pos.x) and feql(p.y, pos.y) then
			table.insert(res, obj)
		end
	end
	if #res == 0 then
		return nil
	else
		return res
	end
end

function room.is_walkable(pos)
	local objs = room.find_at_pos(pos)
	return objs == nil
end

function room.is_opaque(pos)
	local objs = room.find_at_pos(pos)
	if objs and #objs == 1 then
		return objs[1].opaque
	end
	return objs ~= nil
end

function room.pushable(pos)
	local res = {}
	for _,d in pairs(move_offset) do
		local p = pos + d	
		local objs = room.find_at_pos(p)
		if objs ~= nil and #objs == 1 then
			local o = objs[1]
			if o.id == 'brick' or o.id == 'enemy' or o.id == 'player' then
				table.insert(res, p)
			end
		end
	end
	return res
end

-- returns true if moved, false if blocked
function room.move(obj, offset)
	local target_pos = obj.pos + offset
	local back_pos = obj.pos - offset
	local in_front = room.find_at_pos(target_pos)
	local in_back = room.find_at_pos(back_pos)

	local result = true 

	if in_front then
		-- apply > rule
		for i,fobj in ipairs(in_front) do
			local moved = false
			for j,r in ipairs(rules.desc) do
				local d, a, b, dir_a, dir_b = unpack(r)
				if d == '>' and obj.id == a and fobj.id == b then
					if dir_b == '>' then
						if room.move(fobj, offset) then
							if dir_a ~= '>' then
								result = false
							else
								moved = true
							end
						else
							result = false
						end
					else
						if dir_a == '>' then
							moved = true
						else
							result = false
						end
					end
				end
			end
			if not moved then
				result = false
			end
		end
	end

	if result and in_back then
		-- apply < rule
		for _,fobj in ipairs(in_back) do
			for _,r in ipairs(rules.desc) do
				local d, a, b, dir_a, dir_b = unpack(r)
				if d == '<' and obj.id == a and fobj.id == b then
					if dir_b == '<' then
						local old_pos = obj.pos
						obj.pos = target_pos
						room.move(fobj, offset)
						obj.pos = old_pos
					end
				end
			end
		end
	end

	if result then
		obj.next_pos = target_pos
	end

	return result
end

function room.clone_state()
	local objects_clone = {}
	for i,obj in ipairs(objects) do
		table.insert(objects_clone, {
			pos = vec2(obj.pos)
		})
	end

	return objects_clone
end

function room.input(id, action)
	if action == 'pause' then
		states.pop()
	end
end

function room.win_condition()
	local enemy_count = 0
	room.for_each_enemy(function()
		enemy_count = enemy_count + 1
	end)
	return enemy_count == 0
end

function room.lose_condition()
	local unit_count = 0
	room.for_each_unit(function()
		unit_count = unit_count + 1
	end)
	return unit_count == 0
end

function room.update()
	local mouse_pos = mouse.pos()
	local hover_tile = screen2world(mouse_pos)
	
	if room_state ~= 'enemies' then
	
		if hover_tile ~= nil then
			if mouse.down(mouse.primary) and selected_unit then
				local u = selected_unit
				if u.walkable and pos_in_table(hover_tile, u.walkable) then
					-- start unit move animation
					if not eql_pos(hover_tile, u.pos) then
						u.path = bfs.path(u.pos, hover_tile, room.is_walkable)
						u.next_pos = u.path[#u.path]
						u.moved = true
						u.walkable = nil
						table.remove(u.path, #u.path)
						room_state = 'move'
						animation = true
						animation_start = states.time()
						selected_unit.walkable = nil
					end
				end

				if u.pushable and pos_in_table(hover_tile, u.pushable) then
					local offset = hover_tile - u.pos	
					if room.move(selected_unit, offset) then
						mfx.trigger('push')
						u.pushed = true
						u.pushable = nil
						animation = true
						animation_start = states.time()
						selected_unit.pushable = nil
					end
				end
					
				selected_unit = nil
			end

			-- show visibility field for enemy when hovering over it
			local units = room.find_at_pos(hover_tile)
			if units and #units == 1 and units[1].id == 'enemy' then
				ai.update_visibility(units[1], room)
			end

			-- draw selected tile		
			sprsheet.draw_centered(
				'brick', overlay_layer,
				world2screen(hover_tile), 0, 64/48,
				rgba(1, 1, 1, 0.2)
			)

			if mouse.down(mouse.primary) then
				-- select unit
				if units ~= nil and #units == 1 then
					local u = units[1]
					if u.id == 'player' and u.next_pos == nil then
						if not u.moved then
							-- mark walkable tiles
							selected_unit = u
							u.walkable = bfs.walkable(u.pos, room.is_walkable, 4)
						end
						if not u.pushed then
							-- mark pushable objects
							selected_unit = u	
							u.pushable = room.pushable(u.pos)
						end
					end
				end
			end
		end

		if char.down(' ') then
			-- end turn
			room.for_each_unit(function(u)
				u.moved = false
				u.pushed = false
				selected_unit = nil
			end)

			-- move enemies
			room_state = 'enemies'
			room.for_each_enemy(function(e)
				e.moved = false
			end)
		end

		if char.down('r') then
			states.pop()
		end
	else
		local moved_any = false
		if not animation then
			room.for_each_enemy(function(e)
				if not e.moved and not moved_any then
					ai.think(e, room)
					animation = true
					animation_start = states.time()
					e.moved = true
					moved_any = true
				end
			end)

			if not moved_any then
				room_state = 'wait'
			end
		end
	end

	return true
end

function room.fade_alpha(x, y, t)
	return t
end

function room.render_floor(t)
	local alpha = 1
	if t ~= 0 then
		if t < 0 then
			t = -t
		end
		alpha = math.abs(1-t)
	end
	local tint = rgba(1, 1, 1, alpha)
	for y=0,height-1 do
		for x=0,width-1 do
			local pos = world2screen(vec2(x, y))
			sprsheet.draw_centered('floor', 0, pos, tint)
		end
	end
end

function room.render(t)
	room.render_floor(t)

	local anim_t = nil
	if animation then
		anim_t = (states.time() - animation_start) / animation_len
		anim_t = math.min(1, anim_t)
	end
	for i,obj in ipairs(objects) do
		if obj.sprite then
			local is_selected = obj == selected_unit
			local is_enemy = obj.id == 'enemy'
			local pos = world2screen(obj.pos)
			if anim_t and obj.next_pos then
				local next_pos = world2screen(obj.next_pos)
				pos = lerp(pos, next_pos, anim_t)
			end
			pos.x = math.floor(pos.x)
			pos.y = math.floor(pos.y)

			local alpha = 1
			if t ~= 0 then
				if t < 0 then
					t = - t
				end
				alpha = room.fade_alpha(pos.x, pos.y, math.abs(1-t))
			end

			local scale = 1

			-- falling down the hole...
			if obj.id == '' and obj.disappear_t then
				local dt = states.time() - obj.disappear_t
				if dt > 1 then
					dt = 1
				end
				obj.rot = smoothstep(0, math.pi, dt) 
				alpha = smoothstep(1, 0, dt)
				scale = smoothstep(1, 0.2, dt)

				if dt == 1 then
					obj.delete = true
				end
			end

			if obj.id == 'player_die' then
				local t = states.time() - obj.die_t
				local frame = math.floor(t * 10)
				if frame >= 3 then
					obj.delete = true
				else
					obj.sprite = 'player_die'..tostring(frame+1)
				end
			end
		
			-- special logic for selected unit
			if is_selected then
				-- move up and down a bit to make selection visible
				local dy = math.sin(states.time() * 6)^2 * 2
				pos.y = pos.y + dy

				-- render the walkable tiles
				if obj.walkable then
					for _,pos in ipairs(obj.walkable) do
						sprsheet.draw_centered(
						'brick', overlay_layer,
						world2screen(pos), 0, 64/48,
						rgba(1, 1, 1, 0.2)
					)
					end
				end
				if obj.pushable then
					for _,pos in ipairs(obj.pushable) do
						sprsheet.draw_centered(
							'brick', push_overlay_layer,
							world2screen(pos), 0, 64/48,
							rgba(1, 0, 0, 0.2)
						)
					end
				end
			end

			-- special logic for enemy
			if is_enemy then
				-- render the seeable tiles
				if obj.visible then
					for _,pos in ipairs(obj.visible) do
						sprsheet.draw_centered(
							'brick', overlay_layer,
							world2screen(pos), 0, 64/48,
							rgba(0.1, 0.1, 0.8, 0.2)
						)
					end
					obj.visible = nil
				end
			end

			local spr = obj.sprite
			if obj.id == 'enemy' and obj.alarmed then
				spr = 'enemy_red'
			end

			local layer = obj_layer
			if obj.layer then
				layer = layer + obj.layer
			end
			local old_alpha = obj.tint.a
			obj.tint.a = alpha
			sprsheet.draw_centered(
				spr, layer, pos, obj.rot, scale, obj.tint

			)
			obj.tint.a = old_alpha
		end
	end

	-- update objects after animation, check win condition
	local moved = false
	local play_step = false
	if anim_t == 1 then
		for i,obj in ipairs(objects) do
			if obj.next_pos then
				if not eql_pos(obj.pos, obj.next_pos) then
					play_step = true
				end
				obj.pos = obj.next_pos
				obj.old_pos = vec2(obj.pos)
				obj.next_pos = nil
				

				if obj.path ~= nil then
					local n = #obj.path
					obj.next_pos = obj.path[n]
					moved = true
					animation_start = states.time()
					table.remove(obj.path, n)
					if #obj.path == 0 then
						obj.path = nil
					end
				end
			end
		end
		if play_step then
			mfx.trigger('step')
		end

		if moved == false then
			-- make things fall down the holes
			for i,obj in ipairs(objects) do
				if obj.id == 'hole' then
					local objs = room.find_at_pos(obj.pos)
					for j,o in ipairs(objs) do
						-- everything else besides hole falls into hole
						if o.id ~= 'hole' then
							o.id = ''
							o.disappear_t = states.time()
							mfx.trigger('fall')
						end
					end
				end
			end

			-- make things die
			for i,obj in ipairs(objects) do
				if obj.id == 'enemy' then
					local objs = room.find_at_pos(obj.pos)
					for j,o in ipairs(objs) do
						-- enemy kills everything else besides enemy 
						if o.id ~= 'enemy' and not o.die_t then
							if o.id == 'player' then
								o.id = 'player_die'
								room.units_died = room.units_died + 1
								mfx.trigger('die')
							end

							o.die_t = states.time()
						end
					end
				end
			end
			animation = false
		end
	end

	local has_dying_objs = false
	for i,obj in ipairs(objects) do
		if obj.id == '' or obj.id == 'player_die' then
			has_dying_objs = true
			break
		end
	end

	if not has_dying_objs and not room.exiting then
		if room.win_condition() then
			room.win = true
			room.exiting = true
			states.pop()
		end

		if room.lose_condition() then
			states.pop()
			room.exiting = true
		end
	end

	-- delete objects
	local objs = {}
	for i,obj in ipairs(objects) do
		if not obj.delete then
			table.insert(objs, obj)
		end
	end
	objects = objs

	return true
end


return room
